import time
import os
from .encoder import Encoder
from .alphabet import Alphabet

class ShortId:

    def __init__(self):
        self._counter = 0
        self._previous_delta = 0

        self.reduce_delay = 86400
        self.reduce_time = (int(time.time()) * 1000) - self.reduce_delay
        self.version = 4
        self.encoder = Encoder()
        self.alphabet = Alphabet()

    def generate(self):
        delta = self._get_time_delta()

        if delta == self._previous_delta:
            self._counter += 1
        else:
            self._counter = 0
            self._previous_delta = delta

        res = self.encoder.encode(
            self.alphabet.lookup,
            self.version
        )
        res = res + self.encoder.encode(
            self.alphabet.lookup,
            os.getpid()
        )

        if self._counter > 0:
            res = res + self.encoder.encode(
                self.alphabet.lookup,
                self._counter
            )

        res = res + self.encoder.encode(
            self.alphabet.lookup,
            delta
        )

        return res

    def _get_time(self):
        return int(time.time() * 1000)

    def _get_time_delta(self):
        return int(
            round(
                (self._get_time() - self.reduce_time) * 0.01
            )
        )



import string
from random import random
from random import shuffle

class Alphabet:

    ORIGINAL = "{0}{1}_-".format(
        string.digits,
        string.ascii_letters
    )

    def __init__(self):
        self._alphabet = Alphabet.ORIGINAL
        self.reset()

    def set(self, characters):
        characters_len = len(characters)
        original_len = len(Alphabet.ORIGINAL)
        if characters_len != original_len:
            raise RuntimeError(
                "Custom alphabet for shortId must be {0} unique characters. " +
                "You submitted {1} characters: {2}".format(
                    original_len,
                    characters_len,
                    characters
                )
            )

        characters_set_len = len(set(characters))
        if characters_set_len != characters_len:
            raise RuntimeError(
                (
                    "Custom alphabet for shortId must be {0} unique characters. "
                    "{1} non unique values found."
                ).format(characters_len - characters_set_len)
            )

        self._alphabet = characters
        self.reset()

    def get(self):
        return self._alphabet

    def reset(self):
        self._shuffled = self.shuffle()

    def lookup(self, index):
        return self.get_shuffled()[index]

    def shuffle(self):
        alphabet = list(self._alphabet)
        shuffle(alphabet)
        return "".join(alphabet)

    def get_shuffled(self):
        return self._shuffled



import os

class Encoder:
    def encode(self, lookup_fn, number):
        loops = 0
        res = ""

        while True:
            res = res + lookup_fn(
                ((number >> (4 * loops)) & 0x0F
            ) | (
                bytearray(os.urandom(1))[0] & 0x30)
            )
            loops += 1
            if (number < pow(16, loops)) != 0:
                break

        return res



import unittest
from shortid import ShortId

class TestShortId(unittest.TestCase):
    def setUp(self):
        self.shortid = ShortId()

    def test_should_be_unambiquous_on_a_bunch_of_iterations(self):
        ids = []
        for i in range(0, 100000):
            ids.append(self.shortid.generate())

        self.assertEqual(len(set(ids)), len(ids))

    def test_should_be_unambiquous_on_a_bunch_of_iterations_new_instance(self):
        ids = []
        for i in range(0, 100000):
            ids.append(ShortId().generate())

        self.assertEqual(len(set(ids)), len(ids))

    def test_generate_max_length(self):
        lengths = []
        for i in range(0, 50000):
            lengths.append(len(self.shortid.generate()))
        self.assertEqual(max(lengths) < 12, True)


    def test_generate_max_length_new_instance(self):
        lengths = []
        for i in range(0, 50000):
            lengths.append(len(ShortId().generate()))
        self.assertEqual(max(lengths) < 12, True)



#!/usr/bin/env python

from distutils.core import setup

__version__ = "0.1.2"

setup(
    name="shortid",
    version=__version__,
    description="Short id generator",
    author="Dmitry Moskowski",
    author_email="me@corpix.ru",
    url="https://github.com/corpix/shortid",
    download_url="https://github.com/corpix/shortid/archive/{0}.tar.gz".format(__version__),
    keywords = ["short", "id", "uuid", "shortid", "tinyid"],
    packages=["shortid"],
    license="MIT",
    classifiers=[],
)
